from enum import Enum, auto

from sampo.schemas.graph import GraphNode

StageType = tuple[GraphNode, dict[str, GraphNode]]


class SyntheticGraphType(Enum):
    General = auto()
    Parallel = auto()
    Sequential = auto()



from copy import deepcopy
from typing import List, Dict, Tuple, Optional, Set

from sampo.schemas.graph import GraphNode, GraphNodeDict, GraphEdge, WorkGraph, EdgeType
from sampo.schemas.requirements import WorkerReq
from sampo.schemas.works import WorkUnit

STAGE_SEP = '_stage_'


def make_start_id(work_unit_id: str, ind: int) -> str:
    """
    Creates an auxiliary id for restructuring the graph
    :param work_unit_id: str - work unit id
    :param ind: int - sequence number of work_unit stage
    :return:
       auxiliary_id: str - an auxiliary id for the work unit
    """
    return f"{work_unit_id}{STAGE_SEP}{ind}"


def find_lags(edges: List[GraphEdge], edge_type: EdgeType, is_reversed: bool) -> (List[Tuple[float, float, bool]]):
    """
    Searches for the maximum lag among the given edges and type of edge,
        for the maximum lag saves the amount of parental work
    :param is_reversed: bool - Used to specify from which node lag is used
    :param edges: List[GraphEdge] - the given edges
    :param edge_type: str - the given type of edge
    :return:
        max_lag: float - the maximum lag
        parent_volume: float - The volume for which the lag is set, i.e. the parent volume
    """
    lag_volume_list = [(edge.lag, edge.start.work_unit.volume, is_reversed)
                       for edge in edges if edge_type is edge.type]
    return lag_volume_list


def node_restructuring(origin_node: GraphNode, id2new_nodes: GraphNodeDict,
                       lags_volumes_list: List[Tuple[float, float, bool]],
                       old_id_lag2new_id: Dict[Tuple[str, float, bool], str]):
    """
    Splits the node into two parts: into "piece" and "mian" whose sizes are proportional to max_lag and
    parent_volume-max_lag, respectively. The order of "piece" and "mian" is set. For the first vertex the id is changed,
    for the second one it remains the same, so that it is more convenient to restore the edges. The resulting nodes are
    created without edges. It connects two obtained nodes with an unbreakable edge,
    which does not allow to perform tasks in any way
    :param lags_volumes_list: List[Tuple[float, float, bool]] -
    :param origin_node: GraphNode - Node to be divided into two parts
    :param id2new_nodes: GraphNodeDict - Dictionary with restructured new nodes where the restructured nodes will
        be written
    :param old_id_lag2new_id: Dict[Tuple[str, float], str]
    :return: Nothing
    """
    wu = origin_node.work_unit
    if len(lags_volumes_list) == 0 or wu.is_service_unit:
        id2new_nodes[wu.id] = GraphNode(deepcopy(wu), [])
        return

    # last elem - whole work_unit
    proportions_accum = sorted(
        [(1 * int(is_reversed) + lag / volume * (1 - 2 * int(is_reversed)), lag, is_reversed)
         for lag, volume, is_reversed in lags_volumes_list
         if volume > 0 and lag < volume] + [(1, -1, False)])

    proportions_accum.sort()
    proportions: List[Tuple[float, float, bool]] = [proportions_accum[0]]
    for ind in range(1, len(proportions_accum)):
        accum, lag, is_reversed = proportions_accum[ind]
        accum_pred, _, _ = proportions_accum[ind - 1]
        if accum == accum_pred and (ind != len(proportions_accum) - 1):
            continue
        proportions.append((accum - accum_pred, lag, is_reversed))

    for ind in range(len(proportions)):
        piece_div_main, lag, is_reversed = proportions[ind]
        reqs = [WorkerReq(wr.kind, wr.volume * piece_div_main, wr.min_count, wr.max_count)
                for wr in wu.worker_reqs]

        volume = wu.volume * piece_div_main

        new_id = make_start_id(wu.id, ind) if ind < len(proportions) - 1 else wu.id
        new_wu = WorkUnit(new_id, f'{wu.name}{STAGE_SEP}{ind}', reqs, group=wu.group,
                          volume=volume, volume_type=wu.volume_type, display_name=wu.display_name)
        parents = [(id2new_nodes[make_start_id(wu.id, ind - 1)], 0, EdgeType.InseparableFinishStart)] if ind > 0 else []
        id2new_nodes[new_id] = GraphNode(new_wu, parents)
        old_id_lag2new_id[(wu.id, lag, is_reversed)] = new_id
    return


def fill_parents(origin_work_graph: WorkGraph, id2new_nodes: GraphNodeDict,
                 old_id_lag2new_id: Dict[Tuple[str, float, bool], str],
                 use_ffs_separately: bool = False):
    """
    Restores edges in the transformed graph

    :param origin_work_graph: WorkGraph - The original unconverted graph
    :param id2new_nodes: GraphNodeDict -dictionary with transformed vertices,
        if some vertices from the original graph are missing,
        the function converts their copy and writes it into the dictionary
    :param old_id_lag2new_id: Dict[Tuple[str, float, bool], str]
    :param use_ffs_separately:
        If false, then FFS edges are considered equivalent to FS,
        otherwise they are converted as a separate type of edges
    :return: Nothing
    """

    # from the first element since the zero node is the starting node that has no parents
    for node in origin_work_graph.nodes[1:]:
        zero_stage_id = make_start_id(node.id, 0) if make_start_id(node.id, 0) in id2new_nodes else node.id
        last_stage_id = node.id

        parents_zero_stage: List[Tuple[GraphNode, float, EdgeType]] = []
        parents_last_stage: List[Tuple[GraphNode, float, EdgeType]] = []
        for edge in node.edges_to:
            if edge.type in [EdgeType.FinishStart, EdgeType.InseparableFinishStart] or \
                    not use_ffs_separately and edge.type is EdgeType.LagFinishStart:
                parents_zero_stage.append((id2new_nodes[edge.start.id], edge.lag, edge.type))
            elif edge.type is EdgeType.StartStart:
                work_id = make_start_id(edge.start.id, 0)
                work_id = work_id if work_id in id2new_nodes else edge.start.id
                parents_zero_stage.append((id2new_nodes[work_id], edge.lag, EdgeType.FinishStart))
            elif edge.type is EdgeType.FinishFinish:
                parents_last_stage.append((id2new_nodes[edge.start.id], edge.lag, EdgeType.FinishStart))
            elif use_ffs_separately and edge.type is EdgeType.LagFinishStart:
                parents_zero_stage.append(
                    (id2new_nodes[old_id_lag2new_id[edge.start.id, edge.lag, False]], 0, EdgeType.FinishStart))
                id2new_nodes[last_stage_id].add_parents([(id2new_nodes[edge.start.id], 0, EdgeType.FinishStart)])
        id2new_nodes[zero_stage_id].add_parents(parents_zero_stage)
        id2new_nodes[last_stage_id].add_parents(parents_last_stage)

        # add SF connections from origin graph
        parents = [(id2new_nodes[edge.finish.id], edge.lag, EdgeType.FinishStart)
                   for edge in node.edges_from if edge.type is EdgeType.StartFinish]
        # after reversing the start-finish node could remain without a parent, if so, suspend it in the start node
        # parents = parents or [id2new_nodes[origin_work_graph.start.id]]
        id2new_nodes[zero_stage_id].add_parents(parents)

    start_id = origin_work_graph.start.id
    finish_id = origin_work_graph.finish.id
    has_child: Set[str] = {start_id, finish_id}
    has_parent: Set[str] = {start_id, finish_id}
    for node in id2new_nodes.values():
        for edge in node.edges_to:
            has_child.add(edge.start.id)
            has_parent.add(edge.finish.id)

    has_no_child = set(id2new_nodes.keys()) - has_child
    id2new_nodes[origin_work_graph.finish.id].add_parents([id2new_nodes[node_id] for node_id in has_no_child])

    start_id = origin_work_graph.start.id
    has_no_parent = set(id2new_nodes.keys()) - has_parent
    for work_id in has_no_parent:
        id2new_nodes[work_id].add_parents([id2new_nodes[start_id]])


def graph_restructuring(wg: WorkGraph, use_lag_edge_optimization: Optional[bool] = False) -> WorkGraph:
    """
    Rebuilds all edges into finish-start edges with the corresponding rebuilding of the nodes

    :param wg: WorkGraph - The graph to be converted
    :param use_lag_edge_optimization: bool - if true - do optimization fake-finish-start edges,
        otherwise considers such edges to be similar to finish-start
    :return:
        new_work_graph: WorkGraph - restructured graph
    """
    id2new_nodes: GraphNodeDict = dict()
    old_id_lag2new_od: Dict[Tuple[str, float, bool], str] = dict()
    for node in wg.nodes:
        lags_volumes = []
        lags_volumes += find_lags(node.edges_from, EdgeType.StartStart, False)
        lags_volumes += find_lags(node.edges_to, EdgeType.FinishFinish, True)
        if use_lag_edge_optimization:
            lags_volumes += find_lags(node.edges_to, EdgeType.LagFinishStart, True)
            lags_volumes += find_lags(node.edges_from, EdgeType.LagFinishStart, False)
        node_restructuring(node, id2new_nodes, lags_volumes, old_id_lag2new_od)
    fill_parents(wg, id2new_nodes, old_id_lag2new_od, use_ffs_separately=use_lag_edge_optimization)
    return WorkGraph(id2new_nodes[wg.start.id], id2new_nodes[wg.finish.id])



from abc import ABC, abstractmethod
from operator import attrgetter
from typing import List, Dict, Set, Iterable

from sampo.scheduler.timeline.base import Timeline
from sampo.schemas.contractor import WorkerContractorPool, Contractor
from sampo.schemas.graph import GraphNode
from sampo.schemas.requirements import WorkerReq
from sampo.schemas.resources import Worker
from sampo.schemas.schedule import ScheduledWork
from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.utilities.collections_util import build_index

PRIORITY_SHUFFLE_RADIUS = 0.5


class OrderLocalOptimizer(ABC):

    @abstractmethod
    def optimize(self, node_order: list[GraphNode], area: range) -> list[GraphNode]:
        ...


class ScheduleLocalOptimizer(ABC):

    def __init__(self, timeline_type: type[Timeline]):
        """
        :param timeline_type: timeline used for schedule recalculation
        """
        self._timeline_type = timeline_type

    @abstractmethod
    def optimize(self, scheduled_works: dict[GraphNode, ScheduledWork], node_order: list[GraphNode],
                 contractors: list[Contractor], spec: ScheduleSpec, worker_pool: WorkerContractorPool,
                 work_estimator: WorkTimeEstimator, assigned_parent_time: Time,
                 area: range) -> dict[GraphNode, ScheduledWork]:
        """
        Optimizes works `scheduled_works`, referenced by `node_order` and `area` parameters.

        Result writes to `scheduled_works` in-place.
        :param node_order:
        :param contractors:
        :param spec:
        :param worker_pool:
        :param work_estimator:
        :param assigned_parent_time:
        :param scheduled_works:
        :param area:
        """
        ...


def get_swap_candidates(node: GraphNode,
                        node_index: int,
                        candidates: Iterable[GraphNode],
                        node2ind: Dict[GraphNode, int],
                        processed: Set[GraphNode]) -> list[GraphNode]:
    """
    Abstract function to find nodes that can be swapped
    with given node without breaking topological order

    :param node: target node
    :param node_index: index of target node in global sequence
    :param candidates: list of candidates to swapping
    :param node2ind: a dict from node to it's index
    :param processed: a set of nodes that should not be swapped yet
    """
    cur_children: Set[GraphNode] = node.children_set

    def is_candidate_accepted(candidate: GraphNode) -> bool:
        if candidate in cur_children or candidate in processed:
            return False
        candidate_ind = node2ind[candidate]
        for child in cur_children:
            if node2ind.get(child, 0) >= candidate_ind:  # we have a child between us and candidate
                return False
        candidate_parents = candidate.parents
        for parent in candidate_parents:
            if node2ind.get(parent, 0) <= node_index:  # candidate has a parent between us and candidate
                return False
        return True

    return [candidate for candidate in candidates if is_candidate_accepted(candidate)]


class SwapOrderLocalOptimizer(OrderLocalOptimizer):
    """
    This performs just small shuffle that not break topological order
    """

    def optimize(self, node_order: list[GraphNode], area: range) -> list[GraphNode]:
        if node_order is None:
            return

        start_index = area.start
        end_index = area.stop

        # TODO Examine what is better: perform shuffling in nearly placed sub-seq or in whole sequence
        # node2cost = {node: work_priority(node, calculate_working_time_cascade, work_estimator) for node in sub_seq}

        # preprocessing
        node2ind: Dict[GraphNode, int] = {node: start_index + ind for ind, node in
                                          enumerate(node_order[start_index:end_index])}

        # temporary for usability measurement
        swapped = 0

        processed: Set[GraphNode] = set()
        for i in reversed(area):
            node = node_order[i]
            if node in processed:
                continue
            # cur_cost = node2cost[node]
            chain_candidates: List[GraphNode] = node_order[start_index:i]

            accepted_candidates = get_swap_candidates(node, i, chain_candidates, node2ind, processed)

            if accepted_candidates:
                chain_candidate = accepted_candidates[0]
                swap_idx = node2ind[chain_candidate]
                node_order[i], node_order[swap_idx] = node_order[swap_idx], node_order[i]
                # print(f'Swapped {i} and {swap_idx}')
                processed.add(chain_candidate)
                node2ind[chain_candidate] = i
                node2ind[node] = swap_idx
                swapped += 1

            processed.add(node)
        print(f'Swapped {swapped} times!')

        return node_order


class ParallelizeScheduleLocalOptimizer(ScheduleLocalOptimizer):
    """
    This method finds near placed works and turns it to run in parallel.
    It will take effect only if it's launched after scheduling
    """

    def __init__(self, timeline_type: type[Timeline]):
        super().__init__(timeline_type)

    def recalc_schedule(self,
                        node_order: Iterable[GraphNode],
                        contractors: list[Contractor],
                        spec: ScheduleSpec,
                        node2swork: Dict[GraphNode, ScheduledWork],
                        worker_pool: WorkerContractorPool,
                        assigned_parent_time: Time,
                        work_estimator: WorkTimeEstimator = None) -> dict[GraphNode, ScheduledWork]:
        """
        Recalculates duration and start-finish times in the whole given `seq`.
        This will be useful to call after `parallelize_local_sequence` method
        or other methods that can change the appointed set of workers.
        :param node_order: scheduled works to process
        :param contractors:
        :param spec:
        :param node2swork:
        :param worker_pool:
        :param assigned_parent_time:
        :param work_estimator: an optional WorkTimeEstimator object to estimate time of work
        """

        timeline = self._timeline_type(node_order, contractors, worker_pool)
        node2swork_new: Dict[GraphNode, ScheduledWork] = {}

        id2contractor = build_index(contractors, attrgetter('name'))

        for index, node in enumerate(node_order):
            node_schedule = node2swork[node]
            work_spec = spec.get_work_spec(node.id)
            # st = timeline.find_min_start_time(node, node_schedule.workers, node2swork_new)
            # ft = st + node_schedule.get_actual_duration(work_estimator)
            timeline.schedule(index, node, node2swork_new, node_schedule.workers,
                              id2contractor[node_schedule.contractor], None, work_spec.assigned_time,
                              assigned_parent_time, work_estimator)
            # node_schedule.start_end_time = (st, ft)
            node2swork_new[node] = node_schedule

        return node2swork_new

    def optimize(self, scheduled_works: dict[GraphNode, ScheduledWork], node_order: list[GraphNode],
                 contractors: list[Contractor], spec: ScheduleSpec, worker_pool: WorkerContractorPool,
                 work_estimator: WorkTimeEstimator, assigned_parent_time: Time,
                 area: range) -> dict[GraphNode, ScheduledWork]:
        start_index = area.start
        end_index = area.stop
        
        # preprocessing
        node2ind: Dict[GraphNode, int] = {node: start_index + ind for ind, node in
                                          enumerate(node_order[start_index:end_index])}

        processed: Set[GraphNode] = set()

        for i in reversed(area):
            node = node_order[i]
            if node in processed:
                continue
            chain_candidates: List[GraphNode] = node_order[0:i]
            accepted_candidates = get_swap_candidates(node, i, chain_candidates, node2ind, processed)

            my_schedule: ScheduledWork = scheduled_works[node]
            my_workers: Dict[str, Worker] = build_index(my_schedule.workers, attrgetter('name'))
            my_schedule_reqs: Dict[str, WorkerReq] = build_index(my_schedule.work_unit.worker_reqs, attrgetter('kind'))

            new_my_workers = {}

            # now accepted_candidates is a list of nodes that can(according to dependencies) run in parallel
            for candidate in accepted_candidates:
                candidate_schedule = scheduled_works[candidate]

                candidate_schedule_reqs: Dict[str, WorkerReq] = build_index(candidate_schedule.work_unit.worker_reqs,
                                                                            attrgetter('kind'))

                new_candidate_workers: Dict[str, int] = {}

                satisfy = True

                for candidate_worker in candidate_schedule.workers:
                    my_worker = my_workers.get(candidate_worker.name, None)
                    if my_worker is None:  # these two works are not compete for this worker
                        continue

                    need_me = my_workers[candidate_worker.name].count
                    need_candidate = candidate_worker.count

                    total = need_me + need_candidate
                    my_req = my_schedule_reqs[candidate_worker.name]
                    candidate_req = candidate_schedule_reqs[candidate_worker.name]
                    needed_min = my_req.min_count + candidate_req.min_count

                    if needed_min > total:  # these two works can't run in parallel
                        satisfy = False
                        break

                    candidate_worker_count = candidate_req.min_count
                    my_worker_count = my_req.min_count
                    total -= needed_min

                    add_me = min(my_req.max_count, total // 2)
                    add_candidate = min(candidate_req.max_count, total - add_me)

                    my_worker_count += add_me
                    candidate_worker_count += add_candidate

                    new_my_workers[candidate_worker.name] = my_worker_count
                    new_candidate_workers[candidate_worker.name] = candidate_worker_count

                if satisfy:  # replacement found, apply changes and leave candidates bruteforce
                    print(f'Found! {candidate.work_unit.name} {node.work_unit.name}')
                    for worker in my_schedule.workers:
                        worker_count = new_my_workers.get(worker.name, None)
                        if worker_count is not None:
                            worker.count = worker_count
                    for worker in candidate_schedule.workers:
                        worker_count = new_candidate_workers.get(worker.name, None)
                        if worker_count is not None:
                            worker.count = worker_count
                    # candidate_schedule.start_time = my_schedule.start_time
                    break

        return self.recalc_schedule(reversed(node_order), contractors, spec, scheduled_works,
                                    worker_pool, assigned_parent_time, work_estimator)


def optimize_local_sequence(seq: List[GraphNode],
                            start_ind: int,
                            end_ind: int,
                            work_estimator: WorkTimeEstimator = None):
    # TODO Try to find sets of works with nearly same resources and turn in to run in parallel or vice-versa
    pass



from deap.base import Toolbox

from sampo.schemas.scheduled_work import ScheduledWork

native = True
try:
    from native import decodeEvaluationInfo
    from native import evaluate as evaluator
    from native import freeEvaluationInfo
except ImportError:
    print("Can't find native module; switching to default")
    decodeEvaluationInfo = lambda *args: args
    freeEvaluationInfo = lambda *args: args
    evaluator = None
    native = False


from sampo.scheduler.genetic.converter import ChromosomeType
from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import WorkGraph, GraphNode
from sampo.schemas.resources import Worker
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.utilities.collections_util import reverse_dictionary


class NativeWrapper:
    def __init__(self, toolbox: Toolbox, wg: WorkGraph, contractors: list[Contractor], worker_name2index: dict[str, int],
                 worker_pool_indices: dict[int, dict[int, Worker]], time_estimator: WorkTimeEstimator):
        # the outer numeration. Begins with inseparable heads, continuous with tails.
        numeration: dict[int, GraphNode] = {i: node for i, node in
                                            enumerate(filter(lambda node: not node.is_inseparable_son(), wg.nodes))}
        heads_count = len(numeration)
        for i, node in enumerate([node for node in wg.nodes if node.is_inseparable_son()]):
            numeration[heads_count + i] = node
        rev_numeration = reverse_dictionary(numeration)

        self.numeration = numeration
        # for each vertex index store list of parents' indices
        self.parents = [[rev_numeration[p] for p in numeration[index].parents] for index in range(wg.vertex_count)]
        # for each vertex index store list of whole it's inseparable chain indices
        self.inseparables = [[rev_numeration[p] for p in numeration[index].get_inseparable_chain_with_self()]
                             for index in range(wg.vertex_count)]
        # contractors' workers matrix. If contractor can't supply given type of worker, 0 should be passed
        self.workers = [[0 for _ in range(len(worker_name2index))] for _ in contractors]
        for i, contractor in enumerate(contractors):
            for worker in contractor.workers.values():
                self.workers[i][worker_name2index[worker.name]] = worker.count

        self.totalWorksCount = wg.vertex_count
        self.time_estimator = time_estimator
        self.worker_pool_indices = worker_pool_indices
        self._current_chromosomes = None

        if not native:
            def fit(chromosome: ChromosomeType) -> int:
                sworks = toolbox.chromosome_to_schedule(chromosome)[0]
                return max([swork.finish_time for swork in sworks.values()]).value
            self.evaluator = lambda _, chromosomes: [fit(chromosome) for chromosome in chromosomes]
        else:
            self.evaluator = evaluator

        # preparing C++ cache
        self._cache = decodeEvaluationInfo(self, self.parents, self.inseparables, self.workers, self.totalWorksCount)

    def calculate_working_time(self, chromosome_ind: int, team_target: int, work: int) -> int:
        team = self._current_chromosomes[chromosome_ind][1][team_target]
        workers = [self.worker_pool_indices[worker_index][team[len(self.workers[0])]]
                   .copy().with_count(team[worker_index])
                   for worker_index in range(len(self.workers[0]))
                   if team[worker_index] > 0]
        return self.numeration[work].work_unit.estimate_static(workers, self.time_estimator).value

    def evaluate(self, chromosomes: list[ChromosomeType]):
        self._current_chromosomes = chromosomes
        return self.evaluator(self._cache, chromosomes)

    def close(self):
        freeEvaluationInfo(self._cache)



"""
This file contains high-level functions for calling the resource optimization
"""
from enum import Enum, auto
from functools import partial
from typing import Union, Tuple

from sampo.metrics.resources_in_time.binary_search import BinarySearchOptimizationType, BinarySearchOptimizer
from sampo.metrics.resources_in_time.newton_conjugate_gradient import NewtonCGOptimizer
from sampo.scheduler.base import Scheduler
from sampo.schemas.contractor import WorkerContractorPool, Contractor
from sampo.schemas.graph import WorkGraph
from sampo.schemas.time import Time


class ResourceOptimizationType(Enum):
    """
    Represents resource optimization technique.
    WARNING: Only BinarySearch is currently implemented. Do not Use NewtonCG
    """
    BinarySearch = auto()
    NewtonCG = auto()


def apply_binary_optimization(scheduler: Scheduler, work_graph: WorkGraph, deadline: Time, start: str,
                              method: BinarySearchOptimizationType = BinarySearchOptimizationType.ItemByItemFastInit,
                              agents_from_manual_input: WorkerContractorPool = None,
                              dry_resources: bool = False) \
        -> Union[Tuple[Contractor, Time], Tuple[None, None]]:
    """
    Applies resource optimization by means of binary search to the given work graph with the given scheduler
    :param dry_resources:
    :param start:
    :param ksg_info:
    :param agents_from_manual_input:
    :param scheduler:
    :param work_graph:
    :param deadline: Estimated time of end of work
    :param method: Method of optimization. Can be 'fast', 'item-by-item' or 'item-by-item-fast-init'
    :return:
    """
    optimizer = BinarySearchOptimizer(scheduler, method)
    return optimizer.optimize(work_graph, deadline, start, agents_from_manual_input,
                              dry_resources=dry_resources)


def apply_gradient_optimization(scheduler: Scheduler, work_graph: WorkGraph, deadline: Time,
                                start: str,
                                agents_from_manual_input: WorkerContractorPool = None,
                                dry_resources: bool = False) \
        -> Union[Tuple[Contractor, Time], Tuple[None, None]]:
    """
    WARNING: Gradient resource optimization has not been implemented properly
    Applies resource optimization by means of gradient methods to the given work graph with the given scheduler
    :param scheduler:
    :param work_graph:
    :param deadline:
    :param agents_from_manual_input:
    :return:
    """
    raise NotImplementedError('Gradient resource optimization has not been implemented properly')

    optimizer = NewtonCGOptimizer(scheduler)
    return optimizer.optimize(work_graph, deadline, ksg_info, start,
                              agents=agents_from_manual_input, dry_resources=dry_resources)


def apply_resource_optimization(optimization_type: ResourceOptimizationType, scheduler: Scheduler,
                                work_graph: WorkGraph, deadline: Time,
                                start: str,
                                agents_from_manual_input: WorkerContractorPool = None,
                                dry_resources: bool = False) \
        -> Union[Tuple[Contractor, Time], Tuple[None, None]]:
    if optimization_type is ResourceOptimizationType.BinarySearch:
        optimizer = partial(apply_binary_optimization, method=BinarySearchOptimizationType.ItemByItemFastInit)
    elif optimization_type is ResourceOptimizationType.NewtonCG:
        optimizer = apply_gradient_optimization
    else:
        raise Exception(f'Optimization method not implemented: {optimization_type.name}')

    return optimizer(scheduler=scheduler, work_graph=work_graph, deadline=deadline, start=start,
                     agents_from_manual_input=agents_from_manual_input, dry_resources=dry_resources)



from sampo.generator.base import SimpleSynthetic
from sampo.generator.environment.contractor import get_contractor, get_contractor_by_wg
from sampo.generator.pipeline.extension import extend_resources, extend_names
from sampo.generator.pipeline.project import get_start_stage, get_finish_stage, get_small_graph, get_graph



import uuid
from random import Random
from typing import Optional
from uuid import uuid4



# TODO: describe the function (description, parameters, return type)
def uuid_str(rand: Optional[Random] = None) -> str:
    ans = uuid4() if rand is None else uuid.UUID(int=rand.getrandbits(128))
    return str(ans)


# TODO: describe the function (description, parameters, return type)


"""
# TODO: describe the function (description, parameters, return type)
def binary_search(a, x, lo=0, hi=None):
    if hi is None:
        hi = len(a)
    while lo < hi:
        mid = (lo + hi) >> 1
        if a[mid] < x:
            lo = mid + 1
        else:
            hi = mid
    return lo


# TODO: describe the function (description, parameters, return type)
def binary_search_reversed(a, x, lo=0, hi=None):
    if hi is None:
        hi = len(a)
    while lo < hi:
        mid = (lo + hi) >> 1
        if a[mid] > x:
            lo = mid + 1
        else:
            hi = mid
    return lo
"""



from dataclasses import dataclass
from enum import Enum, auto
from typing import Optional, Tuple, Union

import numpy as np

from sampo.metrics.resources_in_time.base import ResourceOptimizer, is_resources_good, init_borders, prepare_answer
from sampo.scheduler.base import Scheduler
from sampo.schemas.contractor import Contractor, WorkerContractorPool, get_worker_contractor_pool
from sampo.schemas.graph import WorkGraph
from sampo.schemas.time import Time


class BinarySearchOptimizationType(Enum):
    Fast = auto()
    ItemByItem = auto()
    ItemByItemFastInit = auto()


@dataclass(frozen=True)
class BinarySearchOptimizer(ResourceOptimizer):
    scheduler: Scheduler
    method: BinarySearchOptimizationType = BinarySearchOptimizationType.ItemByItemFastInit
    worker_factor: Optional[int] = 1000
    max_workers: Optional[int] = 1000

    def optimize(self, wg: WorkGraph, deadline: Time,
                 agents: Optional[WorkerContractorPool] = None,
                 dry_resources: Optional[bool] = True) \
            -> Union[Tuple[Contractor, Time], Tuple[None, None]]:
        if self.method is BinarySearchOptimizationType.Fast:
            optimize_func = union_search
        elif self.method is BinarySearchOptimizationType.ItemByItem:
            optimize_func = item_by_item_search
        elif self.method is BinarySearchOptimizationType.ItemByItemFastInit:
            optimize_func = item_by_item_fast_init_search
        else:
            raise Exception(f'Unknown Binary Search type: {self.method}')
        return optimize_func(wg, self.scheduler, deadline, self.worker_factor, self.max_workers,
                             agents, dry_resources=dry_resources)


def union_search(wg: WorkGraph, scheduler: Scheduler, deadline: Time,
                 worker_factor: Optional[int] = 1000,
                 max_workers: Optional[int] = 1000,
                 right_agents: Optional[WorkerContractorPool] = None,
                 dry_resources: Optional[bool] = True) -> (Contractor, Time) or (None, None):
    left_counts, right_counts, agent_names = \
        init_borders(wg, scheduler, deadline, worker_factor, max_workers, right_agents)
    if right_counts is None:
        return prepare_answer(left_counts, agent_names, wg, scheduler, False)
    if left_counts is None:
        return prepare_answer(right_counts, agent_names, wg, scheduler, False)

    while True:
        mid_counts: np.array = ((left_counts + right_counts) / 2).astype(int)
        if (mid_counts == right_counts).all() or (mid_counts == left_counts).all():
            break
        if is_resources_good(wg, mid_counts, agent_names, scheduler, deadline):
            right_counts = mid_counts
        else:
            left_counts = mid_counts

    return prepare_answer(right_counts, agent_names, wg, scheduler, dry_resources)


def item_by_item_fast_init_search(wg: WorkGraph, scheduler: Scheduler, deadline: Time,
                                  worker_factor: Optional[int] = 1000,
                                  max_workers: Optional[int] = 1000,
                                  right_agents: Optional[WorkerContractorPool] = None,
                                  dry_resources: Optional[bool] = False) -> (Contractor, Time) or (None, None):
    contractor, max_time = union_search(wg, scheduler, deadline,
                                        worker_factor, max_workers, right_agents,
                                        dry_resources=False)
    if contractor is None:
        return None, None
    agents = get_worker_contractor_pool([contractor])
    contractor, max_time = item_by_item_search(wg, scheduler, deadline, worker_factor, max_workers, agents,
                                               dry_resources)
    return contractor, max_time


def item_by_item_search(wg: WorkGraph, scheduler: Scheduler, deadline: Time,
                        worker_factor: Optional[int] = 1000,
                        max_workers: Optional[int] = 1000,
                        right_agents: Optional[WorkerContractorPool] = None,
                        dry_resources: Optional[bool] = True) -> (Contractor, Time) or (None, None):
    left_counts, right_counts, agent_names = \
        init_borders(wg, scheduler, deadline, worker_factor, max_workers, right_agents)
    if right_counts is None:
        return prepare_answer(left_counts, agent_names, wg, scheduler, False)
    if left_counts is None:
        return prepare_answer(right_counts, agent_names, wg, scheduler, True)

    for index in range(len(left_counts)):
        while True:
            mid_counts = right_counts.copy()
            mid_counts[index] = (right_counts[index] + left_counts[index]) / 2
            if mid_counts[index] == right_counts[index] or mid_counts[index] == left_counts[index]:
                break
            if is_resources_good(wg, mid_counts, agent_names, scheduler, deadline):
                right_counts[index] = mid_counts[index]
            else:
                left_counts[index] = mid_counts[index]

    return prepare_answer(right_counts, agent_names, wg, scheduler, dry_resources)



from enum import Enum

from sampo.pipeline.base import InputPipeline
from sampo.pipeline.default import DefaultInputPipeline
from sampo.pipeline.exception import SchedulingPipelineError


class PipelineType(Enum):
    DEFAULT = 0


class SchedulingPipeline:
    @staticmethod
    def create(pipeline_type: PipelineType = PipelineType.DEFAULT) -> InputPipeline:
        match pipeline_type:
            case PipelineType.DEFAULT:
                return DefaultInputPipeline()
            case _:
                raise SchedulingPipelineError('Unknown pipeline type')



from copy import deepcopy
from operator import attrgetter
from typing import Callable

from sampo.scheduler.utils.obstruction import Obstruction
from sampo.schemas.graph import WorkGraph, GraphNode
from sampo.utilities.collections_util import build_index


class BlockNode:
    """
    `BlockNode` represents the node of `BlockGraph` and contains corresponding
    `WorkGraph` and related dependencies between blocks in `BlockGraph`.
    """
    def __init__(self, wg: WorkGraph, obstruction: Obstruction | None = None):
        self.wg = wg
        self.obstruction = obstruction
        self.blocks_from: list[BlockNode] = []
        self.blocks_to: list[BlockNode] = []

    @property
    def id(self):
        return self.wg.start.id

    def __hash__(self):
        return hash(self.id)


class BlockGraph:
    """
    Represents the block graph, where blocks are instances of `WorkGraph` and
    edges are simple *FS* dependencies.
    """
    def __init__(self, nodes: list[BlockNode]):
        self.nodes = nodes
        self.node_dict = build_index(self.nodes, attrgetter('id'))

    @staticmethod
    def pure(nodes: list[WorkGraph], obstruction_getter: Callable[[int], Obstruction | None] = lambda _: None):
        return BlockGraph([BlockNode(node, obstruction_getter(i)) for i, node in enumerate(nodes)])

    def __getitem__(self, item) -> BlockNode:
        return self.node_dict[item]

    def __len__(self):
        return len(self.nodes)

    def to_work_graph(self) -> WorkGraph:
        """
        Creates `WorkGraph` that is equal to this `BlockGraph`.
        """
        copied_nodes = deepcopy(self.nodes)
        for node in copied_nodes:
            # node.wg = WorkGraph._deserialize(node.wg._serialize())
            for wg_node in node.wg.nodes:
                wg_node.__dict__.pop('parents', None)
                wg_node.__dict__.pop('parents_set', None)
                wg_node.__dict__.pop('children', None)
                wg_node.__dict__.pop('children_set', None)
                wg_node.__dict__.pop('inseparable_parent', None)
                wg_node.__dict__.pop('inseparable_son', None)
        global_start: GraphNode = [node for node in copied_nodes if len(node.blocks_to) == 0][0].wg.start
        global_end:   GraphNode = [node for node in copied_nodes if len(node.blocks_from) == 0][0].wg.finish

        for end in copied_nodes:
            end.wg.start.add_parents([start.wg.finish for start in end.blocks_to])

        return WorkGraph(global_start, global_end)

    def toposort(self) -> list[BlockNode]:
        visited = set()
        ans = []

        def dfs(u: BlockNode):
            visited.add(u)
            for v in u.blocks_to:
                if v not in visited:
                    dfs(v)
            ans.append(u)

        for node in self.nodes:
            if node not in visited:
                dfs(node)
        ans.reverse()

        return ans

    @staticmethod
    def add_edge(start: BlockNode, end: BlockNode):
        start.blocks_to.append(end)
        end.blocks_from.append(start)
